package de.fitko.fitconnect.client;

import com.nimbusds.jose.jwk.RSAKey;
import de.fitko.fitconnect.api.domain.auth.OAuthToken;
import de.fitko.fitconnect.api.domain.model.metadata.Metadata;
import de.fitko.fitconnect.api.domain.model.metadata.attachment.Attachment;
import de.fitko.fitconnect.api.domain.model.metadata.data.Data;
import de.fitko.fitconnect.api.domain.model.submission.Submission;
import de.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import de.fitko.fitconnect.api.domain.model.submission.SubmissionSubmit;
import de.fitko.fitconnect.api.domain.validation.ValidationResult;
import de.fitko.fitconnect.api.exceptions.AttachmentUploadException;
import de.fitko.fitconnect.api.exceptions.ClientNotAuthenticatedException;
import de.fitko.fitconnect.api.exceptions.EncryptionException;
import de.fitko.fitconnect.api.exceptions.InvalidPublicKeyException;
import de.fitko.fitconnect.api.services.Sender;
import de.fitko.fitconnect.api.services.Subscriber;

import java.io.File;
import java.nio.charset.StandardCharsets;
import java.util.List;
import java.util.UUID;
import java.util.stream.Collectors;

/**
 * A fluent client for announcing and handing in a {@link SubmissionSubmit}
 */
public class SenderClient {

    private SenderClient() {
    }

    public static WithDestination build(final Sender sender, final String clientId, final String secret) {
        return new ClientBuilder(sender, clientId, secret);
    }

    public interface WithDestination {

        /**
         * Configures the client for the given destination and load the public key
         *
         * @param destinationId unique identifier of the clients destination
         * @return the upload step for attachments
         */
        WithAttachments withDestination(UUID destinationId);

        /**
         * Sends a {@link Submission} with given {@link Attachment}s and json or xml payload to
         * a destinationID.
         *
         * @param destinationId  unique identifier of the clients destination
         * @param attachments list of {@link File} attachments
         * @param data payload as json or xml string
         * @return a {@link SubmissionForPickup} that a {@link Subscriber} receives
         */
        SubmissionForPickup send(UUID destinationId, List<File> attachments, String data);

        /**
         * Sends a {@link Submission} with given {@link Attachment}s and json or xml payload to
         * a destinationID.
         *
         * @param destinationId  unique identifier of the clients destination
         * @param attachments list of {@link File} attachments
         * @param data payload as json or xml byte array
         * @return a {@link SubmissionForPickup} that a {@link Subscriber} receives
         */
        SubmissionForPickup send(UUID destinationId, List<File> attachments, byte[] data);

        /**
         * Sends a {@link Submission} with given {@link Attachment}s to a destinationID.
         *
         * @param destinationId  unique identifier of the clients destination
         * @param attachments list of {@link File} attachments
         * @return a {@link SubmissionForPickup} that a {@link Subscriber} receives
         */
        SubmissionForPickup send(UUID destinationId, List<File> attachments);
    }

    public interface WithAttachments {

        /**
         * Sends the submission with a list of attachments
         *
         * @param attachments that are sent with the submission
         * @return the step where additional data can be added to the submission
         */
        WithData withAttachments(List<File> attachments);
    }

    public interface WithData {

        /**
         * Data as string.
         *
         * @param data json or xml as string
         * @return next step to submit the data
         */
        Submit withData(String data);

        /**
         * Data as byte array.
         *
         * @param data json or xml as byte array
         * @return next step to submit the data
         */
        Submit withData(byte[] data);

        /**
         * Send submission to FIT-Connect API.
         *
         * @return {@link SubmissionForPickup} object of the handed in submission
         */
        SubmissionForPickup submit();
    }

    public interface Submit {
        /**
         * Send submission to FIT-Connect API.
         *
         * @return {@link SubmissionForPickup} object of the handed in submission
         */
        SubmissionForPickup submit();
    }

    public static class ClientBuilder implements WithDestination, WithAttachments, WithData, Submit {

        private final Sender sender;
        private final String clientId;
        private final String secret;

        private UUID destinationId;
        private Data data;
        private List<Attachment> attachments;
        private RSAKey encryptionKey;
        private UUID submissionId;

        public ClientBuilder(final Sender sender, final String clientId, final String secret) {
            this.sender = sender;
            this.clientId = clientId;
            this.secret = secret;
            authenticate();
        }

        @Override
        public WithAttachments withDestination(final UUID destinationId) {
            final RSAKey encryptionKey = sender.getEncryptionKeyForDestination(destinationId);
            final ValidationResult validationResult = sender.validatePublicKey(encryptionKey);
            if (validationResult.hasError()) {
                throw new InvalidPublicKeyException("Public encryption key is not valid", validationResult.getError());
            }
            this.encryptionKey = encryptionKey;
            return this;
        }

        @Override
        public SubmissionForPickup send(final UUID destinationId, final List<File> attachments, final String data) {
            return this.withDestination(destinationId)
                    .withAttachments(attachments)
                    .withData(data)
                    .submit();
        }

        @Override
        public SubmissionForPickup send(final UUID destinationId, final List<File> attachments, final byte[] data) {
            return this.withDestination(destinationId)
                    .withAttachments(attachments)
                    .withData(data)
                    .submit();
        }

        @Override
        public SubmissionForPickup send(final UUID destinationId, final List<File> attachments) {
            return this.withDestination(destinationId)
                    .withAttachments(attachments)
                    .submit();
        }

        @Override
        public Submit withData(final String data) {
            return withData(data.getBytes(StandardCharsets.UTF_8));
        }

        @Override
        public Submit withData(final byte[] data) throws EncryptionException {
            final var unencryptedData = Data.builder().build();
            final Data encryptedData = sender.encryptSubmissionData(encryptionKey, unencryptedData);
            this.data = encryptedData;
            return this;
        }

        @Override
        public WithData withAttachments(final List<File> attachmentFiles) throws AttachmentUploadException {
            final List<Attachment> attachments = readFilesToAttachments(attachmentFiles);
            final SubmissionForPickup submission = createSubmission(destinationId, attachments);
            this.submissionId = submission.getSubmissionId();
            final List<Attachment> encryptedAttachments = encryptAttachments(encryptionKey, attachments);
            sender.uploadAttachments(submissionId, encryptedAttachments);
            return this;
        }

        @Override
        public SubmissionForPickup submit() {
            return sender.sendSubmission(submissionId, getMetadata());
        }

        private Metadata getMetadata() {
            if (this.data == null) {
                return sender.createMetadata(attachments);
            }
            return sender.createMetadata(data, attachments);
        }

        private List<Attachment> encryptAttachments(final RSAKey encryptionKey, final List<Attachment> attachments) {
            return attachments.stream().map(attachment -> encryptAttachment(encryptionKey, attachment)).collect(Collectors.toList());
        }

        private Attachment encryptAttachment(final RSAKey encryptionKey, final Attachment attachment) {
            return sender.encryptAttachment(encryptionKey, attachment);
        }

        private SubmissionForPickup createSubmission(final UUID destinationId, final List<Attachment> attachments) {
            final SubmissionSubmit request = SubmissionSubmit.builder()
                    .destinationId(destinationId)
                    .announcedAttachments(asListOfAttachmentsIds(attachments))
                    .build();

            return sender.createSubmission(request);
        }

        private List<UUID> asListOfAttachmentsIds(final List<Attachment> attachments) {
            return attachments.stream().map(attachment -> attachment.getAttachmentId()).collect(Collectors.toList());
        }

        private void authenticate() {
            final OAuthToken oAuthToken = sender.retrieveOAuthToken(clientId, secret);
            if (oAuthToken == null) {
                throw new ClientNotAuthenticatedException("Client is not authenticated, please authenticate first");
            }
        }

        private List<Attachment> readFilesToAttachments(final List<File> attachmentFiles) {
            return attachmentFiles.stream()
                    .map(this::readFileToAttachment)
                    .collect(Collectors.toList());
        }

        private Attachment readFileToAttachment(final File file) {
            return sender.createAttachment(file);
        }
    }
}